
import java.awt.*;

/**
 * The lunar lander
 *
 * @author YOUR_NAME_HERE
 */
public class LunarLander {

    /**
     * Construct a new lunar lander with the given initial position, velocity, 
     * and angle.  Throttle should be set to 0 initially.  angle should be
     * set to initialAngle, and desiredAngle should be set to the same.
     * Remember to set totalMass and momentum too!  The lander should start
     * out with max fuel.
     * Initialize status to STATUS_FLYING.  
     *
     * @param initialPosition is the initial position of the lander
     * @param initialVelocity is the initial velocity of the lander
     * @param initialAngle    is the initial angle of the lander
     */
    public LunarLander(Vect2D initialPosition, Vect2D initialVelocity,
            double initialAngle) {
        // YOUR CODE HERE
    }

    /**
     * Construct a clone of the given lunar lander.  this LunarLander should 
     * have the same instance variable values as the argument LunarLander 
     * (same position, same momentum, etc.)
     *
     * @param lander is the lunar lander to clone
     */
    public LunarLander(LunarLander lander) {
        // YOUR CODE HERE
    }

    /**
     * Calculate the amount of fuel remaining
     *
     * @return the amount of fuel remaining
     */
    public double getFuel() {
        // YOUR CODE HERE
        return -1;
    }

    /**
     * Get the velocity of the lander
     *
     * @return the velocity of the lunar lander
     */
    public Vect2D getVelocity() {
        // YOUR CODE HERE
        return null;
    }

    /**
     * Advance one timestep.  Remember, if you don't have enough fuel, the rocket
     * should not have any effect!
     *
     * @param dt is the timestep size
     */
    public void step(double dt) {
        // YOUR CODE HERE
    }

    /**
     * Get the status of the lander
     *
     * @return The status of the lunar lander
     */
    public int getStatus() {
        return status;
    }

    /**
     * Get the position of the lander
     *
     * @return the position of the lunar lander
     */
    public Vect2D getPosition() {
        return position;
    }

    /**
     * Get the angle of the lander
     *
     * @return the angle of the lander
     */
    public double getAngle() {
        return angle;
    }

    /**
     * Get the throttle
     *
     * @return the throttle percentage
     */
    public double getThrottle() {
        return throttle;
    }

    /**
     * Set the throttle (between 0 and 1)
     *
     * @param throttlePercentage is the percentage of engine thrust to use
     */
    public void setThrottle(double throttlePercentage) {
        if (getFuel() > 0) {
            throttle = Math.max(0.0, Math.min(1.0, throttlePercentage));
        }
    }

    /**
     * Set the desired angle of the lander.  The lander has a maximum rotation
     * rate, so it may take several timesteps before the lander actually rotates
     * to this angle.
     *
     * @param desiredAngle is the target angle of the lander
     */
    public void setDesiredAngle(double desiredAngle) {
        this.desiredAngle = desiredAngle;
    }

    /**
     * Is the lander touching the ground?
     *
     * @return true if the lander is touching the ground
     */
    private boolean isTouchingGround() {

        // Check each point comprising the lander for collision
        Vect2D[][] landerPolylines = getLanderPolylines();
        for (int i = 0; i < landerPolylines.length; i++) {
            for (int j = 0; j < landerPolylines[i].length; j++) {
                double x = landerPolylines[i][j].getX();
                double y = landerPolylines[i][j].getY();
                if (LunarLanderGame.moon.getTerrainHeight(x) >= y) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Has the lander has landed safely?
     *
     * @return true if the lander has landed safely
     */
    public boolean landedSafely() {
        Moon moon = LunarLanderGame.moon;
        Vect2D[][] landerPolylines = getLanderPolylines();

        // Check if left foot is on the ground
        int leftFootIndex = landerPolylines.length - 1;
        double x = landerPolylines[leftFootIndex][1].getX();
        double y = landerPolylines[leftFootIndex][1].getY();

        if (moon.getTerrainHeight(x) < y) {
            // Left foot is not touching ground
            return false;
        }

        if (!moon.isOverLandingPad(x)) {
            // Left foot hitting plain terrain, not a landing pad
            return false;
        }
        
        // Check if right foot is on the ground
        int rightFootIndex = leftFootIndex - 1;
        x = landerPolylines[rightFootIndex][1].getX();
        y = landerPolylines[rightFootIndex][1].getY();

        if (moon.getTerrainHeight(x) < y) {
            // Right foot is not touching ground
            return false;
        }

        if (!moon.isOverLandingPad(x)) {
            // Right foot hitting plain terrain, not a landing pad
            return false;
        }
        
        // Check if velocities exceeds safe velocities threshold
        if (Math.abs(getVelocity().getY()) > LunarLander.SAFE_VELOCITY_Y ||
                Math.abs(getVelocity().getX()) > LunarLander.SAFE_VELOCITY_X) {
            return false;
        }

        return true;
    }

    /**
     * Get the lunar lander's polylines
     *
     * @return the lunar lander's polylines
     */
    private Vect2D[][] getLanderPolylines() {

        // First, construct the lines which make up the lander.  Center them
        // about (0,0).
        Vect2D[][] landerPolylines = new Vect2D[][]{
            // Cockpit
            {
                new Vect2D(-.37 * LANDER_LENGTH, -.3 * LANDER_LENGTH),
                new Vect2D(-.43 * LANDER_LENGTH, -.18 * LANDER_LENGTH),
                new Vect2D(-.38 * LANDER_LENGTH, 0),
                new Vect2D(-.3 * LANDER_LENGTH, .1 * LANDER_LENGTH),
                new Vect2D(-.1 * LANDER_LENGTH, .2 * LANDER_LENGTH),
                new Vect2D(.1 * LANDER_LENGTH, .2 * LANDER_LENGTH),
                new Vect2D(.3 * LANDER_LENGTH, .1 * LANDER_LENGTH),
                new Vect2D(.38 * LANDER_LENGTH, 0),
                new Vect2D(.43 * LANDER_LENGTH, -.18 * LANDER_LENGTH),
                new Vect2D(.37 * LANDER_LENGTH, -.3 * LANDER_LENGTH)
            },

            // Middle section
            {
                new Vect2D(-.45 * LANDER_LENGTH, -.3 * LANDER_LENGTH),
                new Vect2D(.45 * LANDER_LENGTH, -.3 * LANDER_LENGTH),
                new Vect2D(.45 * LANDER_LENGTH, -.5 * LANDER_LENGTH),
                new Vect2D(-.45 * LANDER_LENGTH, -.5 * LANDER_LENGTH),
                new Vect2D(-.45 * LANDER_LENGTH, -.3 * LANDER_LENGTH)
            },

            // Booster
            {
                new Vect2D(-.2 * LANDER_LENGTH, -.5 * LANDER_LENGTH),
                new Vect2D(-.3 * LANDER_LENGTH, -.75 * LANDER_LENGTH),
                new Vect2D(.3 * LANDER_LENGTH, -.75 * LANDER_LENGTH),
                new Vect2D(.2 * LANDER_LENGTH, -.5 * LANDER_LENGTH)
            },

            // Legs
            {
                new Vect2D(.35 * LANDER_LENGTH, -.5 * LANDER_LENGTH),
                new Vect2D(.45 * LANDER_LENGTH, -.8 * LANDER_LENGTH)
            }, {
                new Vect2D(-.35 * LANDER_LENGTH, -.5 * LANDER_LENGTH),
                new Vect2D(-.45 * LANDER_LENGTH, -.8 * LANDER_LENGTH)
            },

            // Feet
            {
                new Vect2D(.5 * LANDER_LENGTH, -.8 * LANDER_LENGTH),
                new Vect2D(.4 * LANDER_LENGTH, -.8 * LANDER_LENGTH)
            }, {
                new Vect2D(-.5 * LANDER_LENGTH, -.8 * LANDER_LENGTH),
                new Vect2D(-.4 * LANDER_LENGTH, -.8 * LANDER_LENGTH)
            }
        };

        // Now, Let's rotate and translate these lines to put them in the correct location

        // Rotation matrix
        Matrix rotation = new Matrix(Math.cos(angle), -Math.sin(angle),
                Math.sin(angle), Math.cos(angle));

        // Translation vector
        Vect2D translation = new Vect2D(position.getX(), position.getY());

        for (int i = 0; i < landerPolylines.length; i++) {
            for (int j = 0; j < landerPolylines[i].length; j++) {
                landerPolylines[i][j] =
                        translation.add(rotation.times(landerPolylines[i][j]));
            }
        }

        return landerPolylines;
    }

    /**
     * Get the engine flame polylines
     *
     * @return the engine flame polylines
     */
    private Vect2D[] getFlamePolylines() {

        Vect2D[] flamePolylines = new Vect2D[]{
            // Main booster flame
            new Vect2D(-.25 * LANDER_LENGTH, -.75 * LANDER_LENGTH),
            new Vect2D(0, -.75 * LANDER_LENGTH),
            new Vect2D(.25 * LANDER_LENGTH, -.75 * LANDER_LENGTH)
        };

        // Rotation matrix
        Matrix rotation = new Matrix(Math.cos(angle), -Math.sin(angle),
                Math.sin(angle), Math.cos(angle));

        // Translation vector
        Vect2D translation = new Vect2D(position.getX(), position.getY());

        flamePolylines[0] = translation.add(rotation.times(flamePolylines[0]));
        flamePolylines[1] = translation.add(rotation.times(new Vect2D(0, -.75 *
                LANDER_LENGTH - 1.3 * throttle * LANDER_LENGTH)));
        flamePolylines[2] = translation.add(rotation.times(flamePolylines[2]));

        return flamePolylines;
    }

    /**
     * Draw the lunar lander
     *
     * @param g is the graphics context
     */
    public void draw(Graphics g) {
        g.setColor(Color.white);
        // Draw lunar lander
        Vect2D[][] transformedLanderLines = getLanderPolylines();
        for (int i = 0; i < transformedLanderLines.length; i++) {
            drawPolyline(g, transformedLanderLines[i]);
        }

        // Draw main flame
        drawPolyline(g, getFlamePolylines());
    }

    /**
     * Draw the polyline
     *
     * @param g     is the graphics context
     * @param lines is the array of polylines to draw
     */
    private void drawPolyline(Graphics g, Vect2D[] lines) {
        GameCanvas canvas = LunarLanderGame.canvas;

        double worldWidth = canvas.getURX() - canvas.getLLX();
        double worldHeight = canvas.getURY() - canvas.getLLY();
        int canvasWidth = canvas.getWidth();
        int canvasHeight = canvas.getHeight();

        Vect2D translation = new Vect2D(-canvas.getLLX(), -canvas.getLLY());

        for (int i = 0; i < lines.length - 1; i++) {
            // Translate the lines so that the coordinates are offsets from the
            // points (LLX, LLY).
            Vect2D v1 = translation.add(lines[i]);
            Vect2D v2 = translation.add(lines[i + 1]);

            // Convert world coordinates to screen coordinates
            int x1 = (int) ((double) canvasWidth * v1.getX() / worldWidth);
            int y1 = (int) ((double) canvasHeight *
                    (1.0 - (v1.getY() / worldHeight)));
            int x2 = (int) ((double) canvasWidth * v2.getX() / worldWidth);
            int y2 = (int) ((double) canvasHeight *
                    (1.0 - (v2.getY() / worldHeight)));
            g.drawLine(x1, y1, x2, y2);
        }
    }

    // General constants
    public static final double SAFE_VELOCITY_X = 2.0;            // m/s
    public static final double SAFE_VELOCITY_Y = 4.0;            // m/s
    public static final double TURN_RATE = 1.7;                  // rad/s
    public static final double ROTATIONAL_FUEL_BURN_RATE = 90.0; // kg/rad
    public static final double FUEL_BURN_RATE = 200.0;           // kg/s
    public static final double LANDER_LENGTH = 10.0;             // m
    public static final double LANDER_MASS = 35532.0;            // kg
    public static final double MAX_FUEL = 25000.0;               // kg
    public static final double EXHAUST_VELOCITY = 1700.0;        // m/s

    // Status constants
    public static final int STATUS_FLYING = 0;  // Lander is currently flying
    public static final int STATUS_LANDED = 1;  // Lander has safely landed
    public static final int STATUS_CRASHED = 2; // Lander has crashed

    // ROC variables
    private Vect2D position;  // Position of the center of the lander (m, m)
    private Vect2D momentum;  // Mass * velocity of lander (kgm/s, kgm/s)
    private double totalMass; // Mass of lander and fuel (kg)
    
    // Other state variables
    private double throttle;     // Percent engine boost to use
    private double angle;        // Current angle (rad)
    private double desiredAngle; // Target angle (rad)
    private int status;          // Lander's status (flying, crashed, etc.)
}